/* foa.c */


/* =============================================================
   foa: Classifies a color (RGB) image based on the 
        multivariate decision tree represented in the lookup table.
	The result is a binary image with the 'target' regions,
	along with the coordinates for those regions.
 
	Usage is as follows:
	foa lookup_table_file(input) color_image(input) result_image(output)
  
   Shashi Buluswar
   Computer Vision Laboratory
   Dept of Computer Science
   University of Massachusetts
   Amherst, MA  01003
 
   Copyright 1995, University of Massachusetts - Amherst
 
   =============================================================== */

/* Modified: Darrell R. Hougen, February 8, 1995.  Rewrote LUT code to make 
   it more efficient and correct mistakes.  Rewrote the morphing code for the
   same reasons.  Removed ISR library calls.  Made the interface compatible
   with the ICD for Demo C.  Cleaned up some things.  The resulting 
   code is not stand alone.  Instead, it is called from a dummy main
   program or from the RSTA ATR module. */

#include <foa.h>
#include <rsta_types.h>
#include <rsta_modules.h>
#include <stdio.h>

static int read_LUT_from_file(char *fname);
static color_target_detect_LUT_type color_target_detect_LUT;
static int color_target_detect_verbose = 0;

void color_target_detect_set_verbose(int verbose)
{
  color_target_detect_verbose = verbose;
}


int color_target_detect_initialize()
{
  char *path, LUT_path[200];
  int return_code = 0;

  /* Read LUT name. */
  if ((path = getenv("COLOR_LUT")) == NULL)
    {
      if (color_target_detect_verbose)
	fprintf(stderr,
        "ERROR: Environment variable COLOR_LUT not set\n");
      return (1);
    }

  /* Allocate color look up table. */
  color_target_detect_LUT = 
    (color_target_detect_LUT_type)malloc(COLOR_TARGET_DETECT_ARR_SIZE * 
					 sizeof(*color_target_detect_LUT));

  /* Read LUT. */
  sprintf(LUT_path, "%s", path);
  if (color_target_detect_verbose)
    printf ("Reading LUT from file: %s\n", LUT_path);  
  return_code = read_LUT_from_file (LUT_path);
  if (return_code)
    free(color_target_detect_LUT);
  return (return_code);
}

/*=====================================================
  Test for x-(column)overlap.
  =====================================================*/

static int x_overlap(int i, int j,
		       int *beg_x, int *end_x)
{
  if ((i!=j) && (end_x[i]>=end_x[j]) && (end_x[j]>=beg_x[i])) return (1);
  else return(0);
}

/*=====================================================
  Test for y-(row)overlap.
  =====================================================*/

static int y_overlap(int i, int j,
		       int *beg_y, int *end_y)
{
  if ((i!=j) && (end_y[i]>=end_y[j]) && (end_y[j]>=beg_y[i])) return (1);
  else return(0);
}

/*==================================================================
  Check the appropriate neighbor for connected-components rountine
  ==================================================================*/

int checkNeighbor (int *LAB_IMG, int i, int j, int r, int c, int W)
{
  if (LAB_IMG[r*W+c]) {
    LAB_IMG[i*W+j] = LAB_IMG[r*W+c];	    
    return (1);
  } 
  else return (0);
}

/* ====================================================================
   8-neighbor connected components: labels the binary post-classification
   image by connected components.  Currently, the routine uses four passes
   over the image (two top-left to bottom-right, and two the other way).
   ==================================================================== */

int ConnComp (int *I, int w, int h, 
	      int *ret_beg_x, int *ret_beg_y, int *ret_end_x, int *ret_end_y,
	      float TGT_PIX_PCNT) {
  
  int i, j, *LAB_IMG, *TMP_IMG, *THRESH_IMG;
  int num_regions=0, reg_num, *beg_x, *beg_y, *end_x, *end_y, reg_size, ret_num_regions=0;
  int loop=0;
  int min_tgt_pix, tgt_count;
  int rowsum;

  THRESH_IMG = (int *) malloc(sizeof(int)*(w*h));
  LAB_IMG = (int *) malloc(sizeof(int)*(w*h));
  TMP_IMG = (int *) malloc(sizeof(int)*(w*h));
  beg_y = (int *) malloc(sizeof(int)* COLOR_TARGET_DETECT_MAX_DETECTS );
  beg_x = (int *) malloc(sizeof(int)* COLOR_TARGET_DETECT_MAX_DETECTS );
  end_y = (int *) malloc(sizeof(int)* COLOR_TARGET_DETECT_MAX_DETECTS );
  end_x = (int *) malloc(sizeof(int)* COLOR_TARGET_DETECT_MAX_DETECTS );

  min_tgt_pix = (TGT_PIX_PCNT * COLOR_TARGET_DETECT_MIN_REG_SIZE);  /* determine min
								       tgt pixel count */

  if (color_target_detect_verbose)
    printf ("Looking for targets of min size: %d pixels\n", min_tgt_pix);
  

  /* --- Get thresholded image of on-pixels withing tgt_box using sum-array --- */

  /* first element */
  TMP_IMG[0] = I[0];

  /* rest of first row */
  for (j=1; j<w; j++)
    TMP_IMG[j] = I[j] + TMP_IMG[j-1];

  /* rest of the array */
  for (i=1; i<h; i++) {
    rowsum=0;
    for (j=0; j<w; j++) {
      rowsum+= I[i*w+j];
      TMP_IMG[i*w+j] = rowsum + TMP_IMG[(i-1)*w+j];
    }
  }

  for (i=0; i<h; i++) {
    for (j=0; j<w; j++) {
      if ((i>0) && (j>0) && (i<=h-COLOR_TARGET_DETECT_MIN_REG_Y) && 
	  (j<=w-COLOR_TARGET_DETECT_MIN_REG_X)) 
	tgt_count = (TMP_IMG[(i+COLOR_TARGET_DETECT_MIN_REG_Y-1)*w+
			     (j+COLOR_TARGET_DETECT_MIN_REG_X-1)] - 
		     TMP_IMG[(i+COLOR_TARGET_DETECT_MIN_REG_Y-1)*w + (j-1)] - 
		     TMP_IMG[(i-1)*w+(j+COLOR_TARGET_DETECT_MIN_REG_X-1)] + 
		     TMP_IMG[(i-1)*w+(j-1)]);
      else tgt_count = I[i*w+j];

      if (tgt_count > min_tgt_pix) THRESH_IMG[i*w+j] = 1;
      else THRESH_IMG[i*w+j] = 0;

    }
  }

  /* ------------ done generating thresholded image -------------------- */
  /* ------------ now do connected components on thresholded image ----- */
  /* ------------ and generate the labelled image (LAB_IMG) ------------ */
  for (loop=0; loop<2; loop++) {
    
    /* forward scan */
    for (i=1; i<h; i++) {
      for (j=1; j<w; j++) {
	LAB_IMG[i*w+j]=0;
	if (THRESH_IMG[i*w+j]) {
	  if (loop%2 == 0) {
	    if (checkNeighbor (LAB_IMG, i, j, i-1, j, w));        /* north neighbor */ 
	    else if (checkNeighbor (LAB_IMG, i, j, i-1, j-1, w));  /* north-west neighbor */ 
	    else if (checkNeighbor (LAB_IMG, i, j, i, j-1, w));   /* west neighbor */ 
	    else if (checkNeighbor (LAB_IMG, i, j, i+1, j-1, w)); /* south-west neighbor */ 
	    /* new region */
	    else if (loop==0){
	      num_regions++;
	      beg_x[num_regions]=j; beg_y[num_regions]=i; 
	      end_x[num_regions]=j; end_y[num_regions]=i;
	      LAB_IMG[i*w+j] = num_regions; 
	    }

	  }
	  else {
	    if (checkNeighbor (LAB_IMG, i, j, i+1, j- 1, w));     /* south-west neighbor */ 
	    else if (checkNeighbor (LAB_IMG, i, j, i, j-1, w));   /* west neighbor */ 
	    else if (checkNeighbor (LAB_IMG, i, j, i-1, j-1, w)); /* north-west neighbor */ 
	    else if (checkNeighbor (LAB_IMG, i, j, i-1, j, w));   /* north neighbor */ 
	  }
	}
      }
    }
    
    /* backward scan */
    for (i=h-1; i>0; i--) {
      for (j=w-1; j>0; j--) {
	if (LAB_IMG[i*w+j]) {
	  if (loop%2==0) {
	    if (checkNeighbor (LAB_IMG, i, j, i+1, j, w));          /* south neighbor */
	    else if (checkNeighbor (LAB_IMG, i, j, i+1, j+1, w));   /* south-east neighbor */
	    else if (checkNeighbor (LAB_IMG, i, j, i, j+1, w));     /* east neighbor */
	    else if (checkNeighbor (LAB_IMG, i, j, i-1, j+1, w));   /* north-east neighbor */
	  }
	  else {
	    if (checkNeighbor (LAB_IMG, i, j, i-1, j+1, w));        /* north-east neighbor */
	    else if (checkNeighbor (LAB_IMG, i, j, i, j+1, w));     /* east neighbor */
	    else if (checkNeighbor (LAB_IMG, i, j, i+1, j+1, w));   /* south-east neighbor */
	    else if (checkNeighbor (LAB_IMG, i, j, i+1, j, w));     /* south neighbor */
	  }
	}
	
      }
    }
    
  }
  
  /* ------ get bounding boxes from labelled image ------ */
  
  for (i=1; i<h; i++) {
    for (j=1; j<w; j++) {
      reg_num = LAB_IMG[i*w+j];
      if (reg_num) {
	if (j < beg_x[reg_num]) beg_x[reg_num]=j;
	if (i < beg_y[reg_num]) beg_y[reg_num]=i;
	if (j > end_x[reg_num]) end_x[reg_num]=j;
	if (i > end_y[reg_num]) end_y[reg_num]=i;
      }
    }
  }

  /* ----------- get rid of spurious target boxes ------------- */
  
  for (i=0; i<num_regions; i++) {
    reg_size = (end_x[i]-beg_x[i])*(end_y[i]-beg_y[i]);
    if ((reg_size > COLOR_TARGET_DETECT_MIN_REG_SIZE) &&
	(reg_size < COLOR_TARGET_DETECT_MAX_REG_SIZE) &&
	((end_x[i]-beg_x[i]) > COLOR_TARGET_DETECT_MIN_REG_X) &&
	((end_y[i]-beg_y[i]) < COLOR_TARGET_DETECT_MAX_REG_Y) &&
	((end_y[i]-beg_y[i]) > COLOR_TARGET_DETECT_MIN_REG_Y) &&
	((end_y[i]-beg_y[i]) < COLOR_TARGET_DETECT_MAX_REG_Y)) {

      ret_beg_x[ret_num_regions] = beg_x[i];
      ret_end_x[ret_num_regions] = end_x[i];
      ret_beg_y[ret_num_regions] = beg_y[i];
      ret_end_y[ret_num_regions] = end_y[i];
      ret_num_regions++;
    }
  }

  return (ret_num_regions);
  
}

/* =================================================================
   ******
   ****** THIS IS THE OLD CODE for extracting tgt boxes -- 
   ****** I'm leaving this around in case you prefer this 
   ******

   Extract Target Box:
      March across the image with a rectangle of size   
      (COLOR_TARGET_DETECT_MIN_REG_X x COLOR_TARGET_DETECT_MIN_REG_Y)
      and mark off any boxes with >= (COLOR_TARGET_DETECT_MIN_TGT_PIX).

      Then, coalsece overlapping rectangles and eliminate those which
      don't fit the MAX reg_size, reg_x and reg_y criteria.
      
      NOTE: no explicit morhp operations are needed in this method.

   ================================================================== */

static int ExtractTgtBox(int *I, 
			 int w, int h,
			 int *ret_beg_x, int *ret_beg_y, 
			 int *ret_end_x, int *ret_end_y,
			 float TGT_PIX_PCNT)
{
  int i, j, tgt_count, x, rowsum;
  int num_regions = 0;      /* # regions in the binary classified image */
  int *beg_y, *beg_x, *end_y, *end_x, ctr_tmp;
  int *TMP_IMG;
  int reg_size, reg_x, reg_y;
  int min_tgt_pix;         /* threshold for # pixels in a `real' region */

  TMP_IMG = (int *) malloc(sizeof(int)*(w*h));
  beg_y = (int *) malloc(sizeof(int)* COLOR_TARGET_DETECT_MAX_DETECTS );
  beg_x = (int *) malloc(sizeof(int)* COLOR_TARGET_DETECT_MAX_DETECTS );
  end_y = (int *) malloc(sizeof(int)* COLOR_TARGET_DETECT_MAX_DETECTS );
  end_x = (int *) malloc(sizeof(int)* COLOR_TARGET_DETECT_MAX_DETECTS );

  min_tgt_pix = (TGT_PIX_PCNT * COLOR_TARGET_DETECT_MIN_REG_SIZE);


  if (color_target_detect_verbose)
    printf ("Looking for targets of min size: %d pixels\n", min_tgt_pix);
  

  /* --- Get grayscale image of on-pixels withing tgt_box --- */

  /* first element */
  TMP_IMG[0] = I[0];

  /* rest of first row */
  for (j=1; j<w; j++)
    TMP_IMG[j] = I[j] + TMP_IMG[j-1];

  /* rest of the array */
  for (i=1; i<h; i++) {
    rowsum=0;
    for (j=0; j<w; j++) {
      rowsum+= I[i*w+j];
      TMP_IMG[i*w+j] = rowsum + TMP_IMG[(i-1)*w+j];
    }
  }

  for (i=0; i<h; i++) {
    for (j=0; j<w; j++) {
      if ((i>0) && (j>0) && (i<=h-COLOR_TARGET_DETECT_MIN_REG_Y) && 
	  (j<=w-COLOR_TARGET_DETECT_MIN_REG_X)) 
	tgt_count = (TMP_IMG[(i+COLOR_TARGET_DETECT_MIN_REG_Y-1)*w+
			     (j+COLOR_TARGET_DETECT_MIN_REG_X-1)] - 
		     TMP_IMG[(i+COLOR_TARGET_DETECT_MIN_REG_Y-1)*w + (j-1)] - 
		     TMP_IMG[(i-1)*w+(j+COLOR_TARGET_DETECT_MIN_REG_X-1)] + 
		     TMP_IMG[(i-1)*w+(j-1)]);
      else tgt_count = I[i*w+j];

      if (tgt_count > min_tgt_pix) {
	beg_x[num_regions]=j; 
	beg_y[num_regions]=i;
	end_x[num_regions]=j + COLOR_TARGET_DETECT_MIN_REG_X; 
	end_y[num_regions]=i + COLOR_TARGET_DETECT_MIN_REG_Y;
	num_regions++;
      }
    }
  }

  /* === check for overlapping rectangles and combine them === */
  /* === first combine all overlapping rectangles, and     === */
  /* === then eliminate the redundant ones                 === */

  if (color_target_detect_verbose)
    printf ("Checking for overlapping regions...\n");

  for (i=0; i<num_regions; i++) {
    for (j=0; j<num_regions; j++) { 

      if ((x_overlap(i, j, beg_x, end_x) && y_overlap(i, j, beg_y, end_y)) ||
	  (x_overlap(i, j, beg_x, end_x) && y_overlap(j, i, beg_y, end_y)) ||
	  (x_overlap(j, i, beg_x, end_x) && y_overlap(i, j, beg_y, end_y)) ||
	  (x_overlap(j, i, beg_x, end_x) && y_overlap(j, i, beg_y, end_y))) {
	
	if (beg_x[i] < beg_x[j])  beg_x[j]=0;
	else {
	  beg_x[i]=beg_x[j];
	  beg_x[j]=0;
	}
	if (end_x[i] > end_x[j])  end_x[j]=0;
	else {
	  end_x[i]=end_x[j];
	  end_x[j]=0;
	}
	if (beg_y[i] < beg_y[j])  beg_y[j]=0;
	else {
	  beg_y[i]=beg_y[j];
	  beg_y[j]=0;
	}
	if (end_y[i] > end_y[j])  end_y[j]=0;
	else {
	  end_y[i]=end_y[j];
	  end_y[j]=0;
	}
      }
      
    }
  }

  /* remove zeroed out and other invalid regions */
  ctr_tmp=0;
  for (x=0; x<num_regions; x++) {
    reg_x = end_y[x]-beg_y[x];
    reg_y = end_x[x]-beg_x[x];
    reg_size = reg_x*reg_y;
    if ((reg_size != 0) && 
	(reg_size <= COLOR_TARGET_DETECT_MAX_REG_SIZE) &&
	(reg_x <= COLOR_TARGET_DETECT_MAX_REG_X) && 
	(reg_y <= COLOR_TARGET_DETECT_MAX_REG_Y)) {
      ret_beg_y[ctr_tmp]=beg_y[x];
      ret_beg_x[ctr_tmp]=beg_x[x];
      ret_end_y[ctr_tmp]=end_y[x];
      ret_end_x[ctr_tmp]=end_x[x];
      ctr_tmp++;
    }
  }
  num_regions = ctr_tmp;

  return (num_regions);

}

/* =====================================================
   Load a color image and Pixelwise classify it.
   ===================================================== */

int color_target_detect_execute(struct xvimage            *image,
				/* status_sensor_state_msg   *sensor */
				float                      hfov,
				float                      vfov,
				float                      range,
				environ_t                 *environ,
				int                       *num_detects,
				detection_desc_t         **detects,
				float                      TGT_PIX_PCNT)
{
  int i, w, h, n, ntarg = 0;
  unsigned char *R, *G, *B, *r, *g, *b;
  int iR, iG, iB, PixClass, lut_idx, byte_idx;
  int LUT_dec, *I, *ip;
  int *beg_x, *beg_y, *end_x, *end_y; /* coordiates for target regions */
  struct xvimage *result_img, *createimage(), *writeimage();

  if (color_target_detect_verbose)
    printf ("\nStarting color target detection\n");

  w = image->row_size; h = image->col_size;
  n = w * h;
  I = (int *)malloc(n * sizeof(int));

  if (color_target_detect_verbose)
    printf ("Starting pixel classification...\n");

  /* create output binary image */
  result_img = createimage (image->col_size, image->row_size, VFF_TYP_1_BYTE, 1, 1, 
			    "", 0, 0, VFF_MS_NONE,VFF_MAPTYP_NONE, VFF_LOC_IMPLICIT,0);
  
  /* ========================================================
     get pixel RGB
     if it's outside the range, class = 0
     if it's within range, look it up.
     ======================================================== */

  R = (unsigned char *)image->imagedata;
  G = R + n;
  B = G + n;
  ip = I;
  for (i = 0, r = R, g = G, b = B; i < n; i++, r++, g++, b++)
    {
      iR = *r; iG = *g; iB = *b;

#if 0
      PixClass = 0;
      if ((iR / ((iG + iB) / 2.0 + 1.0)) > 1.2)
	{
	  PixClass = 1;
	  ntarg++;
	}
#endif
      
      if ((iR <= COLOR_TARGET_DETECT_RGB_MIN) || 
	  (iR >= COLOR_TARGET_DETECT_RGB_MAX) ||
	  (iG <= COLOR_TARGET_DETECT_RGB_MIN) || 
	  (iG >= COLOR_TARGET_DETECT_RGB_MAX) ||
	  (iB <= COLOR_TARGET_DETECT_RGB_MIN) || 
	  (iB >= COLOR_TARGET_DETECT_RGB_MAX)) 
	PixClass = 0;
      else 
	{
	  iR = iR - COLOR_TARGET_DETECT_RGB_MIN;
	  iG = iG - COLOR_TARGET_DETECT_RGB_MIN;
	  iB = iB - COLOR_TARGET_DETECT_RGB_MIN;

	  lut_idx = ((iR * COLOR_TARGET_DETECT_RANGE + iG) *
		      COLOR_TARGET_DETECT_RANGE + iB);
	  byte_idx = lut_idx & 7;
	  lut_idx >>= 3;

	  LUT_dec = color_target_detect_LUT[lut_idx];

	  if (LUT_dec == 0) PixClass = 0;
	  else if (LUT_dec == 255) PixClass = 1;
	  else PixClass = (LUT_dec >> byte_idx) & 1;
	  if (PixClass == 1) ntarg++;
	  result_img->imagedata[i] = PixClass*128; 

	}
      *ip++ = PixClass;
    }

  writeimage (RESULT_FNAME, result_img);

  if (color_target_detect_verbose == 2)
    printf("  Number of target pixels = %d\n", ntarg);

  if (color_target_detect_verbose)
    printf ( "DONE with pixel classification\n");

  /* === extract the bounding rectangles for target regions === */
  if (color_target_detect_verbose)
    printf ("Extracting bounding rectangles\n");

  beg_x = (int *) malloc(COLOR_TARGET_DETECT_MAX_DETECTS * sizeof(int));
  beg_y = (int *) malloc(COLOR_TARGET_DETECT_MAX_DETECTS * sizeof(int));
  end_x = (int *) malloc(COLOR_TARGET_DETECT_MAX_DETECTS * sizeof(int));
  end_y = (int *) malloc(COLOR_TARGET_DETECT_MAX_DETECTS * sizeof(int));

#if 0
  /* ---- the old method of extracting boxes by marching across the image ----
     ---- with a box of size TGT_MIN_SIZE                                 ---- */
  *num_detects = ExtractTgtBox(I, w, h, beg_x, beg_y, end_x, end_y, TGT_PIX_PCNT);
#endif

#if 1
  *num_detects = ConnComp (I, w, h,  beg_x, beg_y, end_x, end_y, TGT_PIX_PCNT);
#endif

  if (color_target_detect_verbose)
    printf ("ALL DONE...\n");
  if (color_target_detect_verbose)
    printf ("  Number of color detects = %d\n", *num_detects);

  /* Fill in detection structures. */
  *detects = (detection_desc_t *)malloc(*num_detects * sizeof(**detects));
  for (i = 0; i < *num_detects; i++)
    {
      /*printf ("%dx%d - %dx%d\n", beg_x[i], beg_y[i], end_x[i], end_y[i]);*/
      (*detects)[i].result_code = TARGET_DETECTED;
      (*detects)[i].confidence = COLOR_TARGET_DETECT_CONFIDENCE; /* Can we do
								    better? */
      (*detects)[i].CCD_loc.x = beg_x[i];  /* column */
      (*detects)[i].CCD_loc.y = beg_y[i];  /* row    */
      (*detects)[i].CCD_size.x = end_x[i] - beg_x[i] + 1;
      (*detects)[i].CCD_size.y = end_y[i] - beg_y[i] + 1;
      (*detects)[i].CCD_centroid.x = beg_x[i] + (end_x[i] - beg_x[i]) / 2; 
      (*detects)[i].CCD_centroid.y = beg_y[i] + (end_y[i] - beg_y[i]) / 2;
      (*detects)[i].CCD_area = ((end_x[i] - beg_x[i] + 1) * 
				(end_y[i] - beg_y[i] + 1));
      printf ("%dx%d\n", (*detects)[i].CCD_centroid.x, (*detects)[i].CCD_centroid.y);
    }
  
  /* Any memory allocated in this routine should either be freed here
     or in the termination routine below. */

  free(beg_x);  free(beg_y);
  free(end_x);  free(end_y);
  free(I);
  return (0);

}


int color_target_detect_finish()
{
  /* Free lookup table. */
  free(color_target_detect_LUT);
  return (0);
}


/* =========================================================
   Load the lookup table from LUT_file and store in LUT[][][]
   ========================================================== */

static int read_LUT_from_file (char *file_name) 
{
  FILE *LUT_file_ptr;
  
  if ((LUT_file_ptr = fopen (file_name, "r")) == NULL) {
    if (color_target_detect_verbose)
      fprintf (stderr, "Error opening %s\n", file_name);
    return (1);
  }
  else {
    fread (color_target_detect_LUT, sizeof(char), 
	   COLOR_TARGET_DETECT_ARR_SIZE, LUT_file_ptr);
    return(0);
  }
}


