
/*******************************************************************************

Tips and comments for easy integration into the SSV RSTA package:

1.  Execute functions should not read or write files.  The initialize routine
    is used to read in data.  

2.  Only static global variables are permitted.  Global variables are frowned
    upon in general, but static global variables probably need to be used
    (for the verbosity level, for example).  (If the code is written in C++,
    no globals would be required at all.)

3.  No code should call exit().  This is not permitted under any circumstances.
    If your code is baffled, it should return a non-zero value to signify an
    error.  Return a really big number if your code is really baffled.
    The caller will be checking return values in all cases.

4.  Delivered code should include a standalone test program, that calls all
    of the above functions (initialize, execute, finish, set_verbose).

5.  Delivered code should come with a makefile, and the makefile should be
    reasonably clean.  Ideally, the makefile will compile the set of 
    routines into a library that the caller will link against.

6.  GNU C/C++ is the preferred compiler.  (If the code is ansi-C, gnu usually
    doesn't have any trouble with it.)

7.  All print statements within the code should be governed by the verbosity
    level.  The default verbosity level is 0.

    level            meaning
    -----            -------
      0              no print statements.  
      1              print statements on errors, minor (or no) amount of status.
      2              print statements on errors, large amounts of status.

8.  Most ATR algorithms utilize some number of tune-able parameters.  These 
    parameters should be kept in a file and be read in when the initialize
    function is called.  This will make algorithm tuning practical out in the
    field, when working with live imagery.

9.  As described with each initialize function, a single environment variable
    is used for each set of routines.  There are several options on how to 
    use this environment variable:
    If a single data file is needed, 
        env_variable = data_file
    If several data files are needed, 
        env_variable = directory name that contains the several data files.
	    -OR-
        env_variable = file name that contains the several data files.
	    -OR-
        env_variable = colon separated string of names.
    If a single data value were needed (can anyone claim that?), the
        env_variable = the data value             

    Here is a code fragment that uses an environment variable:
    ...
    if( (path = getenv( "RSTA_DAPCODE_DIR" )) == NULL ) {
        if( verbose_flag > 0 )
            fprintf( stderr, 
	             "ERROR:  RSTA_DAPCODE_DIR env variable not set.\n" );
        return -1;
    }
    sprintf( buf, "%s/double_window_probe", path );
    if( (ret=dapcon( buf )) ) {
        if( verbose_flag > 0 )
            printf("TDInitializeDap, ERROR:  dapcon( \"%s\" ) returned %d.\n",
                   buf, ret );
        return(ret);
    }
    ...

10. Environmental data:  The SSV does not currently have temperature and
    humidity sensors which some ATR algorithms can use.  The data structure
    environ_t is defined as a placeholder for additional stuff as it becomes
    available.  Currently environ_t only has fields for the date and time.
    Also:  an environ_t data structure is passed to all execute routines
    defined here, regardless of whether the particular groups stated they
    needed this sort of information.

11. This document is serving as the Interface Control Document (ICD) for
    the routines discussed, and for several others to be added.  As is
    probably obvious, this document is actual C code.  This document is
    the declaration file "rsta_modules.h"  By having the ICD and the 
    declaration file be one and the same, the ICD never gets out of date
    (ICDs often have a tendency to do that).  
    
12. It is planned that a C interface to the elevation map routines will be 
    made available to those algorithms that can use it (currently it is
    in C++).  For those routines, the appropriate imaging geometry will 
    be passed in along with the image data.

13. viff.h is a khoros header file that defines 'struct xvimage'.  Other
    supporting data structures are defined in the file rsta_types.h,
    those data structures are described in that file.

14. Data structures defined in rsta_types.h can be modified to incorporate
    additional needed stuff.  Also, additional data structures can be
    defined and added to that file as needed.  

15. The above names used are not very interesting: there are no organization
    names incorporated into the function names, there are no descriptive
    phrases about algorithms incorporated into the function names.  If two
    algorithms are generally doing the same thing they are given numbers 
    (1, 2..).  If we (later on) get another approach to an unnumbered name, 
    we will change the name to be numbered.  

******************************************************************************/

#include "viff.h"              /* khoros image format definition */
#include "rsta_types.h"        /* supporting data structures     */


/******************************************************************************

	Color Target Detection Routines
	POC:  Ross Beveridge, Colorado State University

******************************************************************************/


/********************************************************************/
#ifdef __cplusplus
extern "C" {
#endif
extern int
color_target_detect_initialize();

/*
 Description:
	Perform initialization.  Read in required data
	from file and set up any required data structures
	to support execution of algorithm, including
        file containing look-up table (or decision tree) 
	to support classification.
	If previously initialized, perform a reset, meaning
	current allocated data is freed, and any needed data files
	are read.

 Inputs:
	none

 Outputs:
	none

 Function Returns:
	0 		on success
	non-zero	on failure (specific error codes, if any, TBD)

 Notes:
	A single environment variable is used to communicate data file 
	name information to this set of routines:
	COLOR_LUT
*/


/********************************************************************/
extern int 
color_target_detect_execute(struct xvimage	     *image,
/* 			    status_sensor_state_msg  *sensor, */
			    float                     hfov,
			    float                     vfov,
			    float 		      range,
			    environ_t                *environ,
			    int			     *num_detects,
			    detection_desc_t        **detects,
			    float                       TGT_PIX_PCNT
			   );

/*
 Description:
	Execute color target detection algorithm.  Provide locations
	(and sizes) within the image estimated to contain targets.

 Inputs:
	struct xvimage   *image;		Color image.
						khoros format.  stored as 3 
						separate bands (r, g, b).

        status_sensor_state_msg                 camera parameters for the color
      	                 *sensor                image, including FOV, etc.

	float	          range;		estimated range to image center
	environ_t        *environ;		currently, the date and time.

 Outputs:
	int 		 *num_detects;	  	the number of detections found

	detection_desc_t **detects;		array of data structures,
						allocated within
						color_target_detect_execute().
						Minimally, the following fields
						are filled out:

	result_code_t    result_code;		with value TARGET_DETECTED

	float		 confidence;		[0..1]; 

    	image_loc_t      CCD_loc;           	upper left location of 
						the bounding box for
						the object.
	image_size_t     CCD_size;          	width and height of the
						bounding box.

	Note:  In the case of no detections, no detection_desc_t's are allocated
	by color_target_detect_execute().

 Function Returns:
	0 		on success
	non-zero	on failure (specific error codes, if any, TBD)

 Notes:  
	time of day is obtained internal to this routine, 
	using unix system call.
*/


/********************************************************************/
extern int 
color_target_detect_finish();

/*
 Description:
	Free up any allocated space used by this routine; 
        return to the pre-initialized state.

 Inputs:
	none

 Outputs:
	none

 Function Returns:
	0 		on success
	non-zero	on failure (specific error codes, if any, TBD)

 Notes:
	none
*/


/********************************************************************/
extern void
color_target_detect_set_verbose(int verbose_level);

/*
 Description:
        set the verbose level to the indicated level.
	level = 0:  No printed output by this module;
	level = 1:  minor descriptive info printed, and error messages printed;
	level = 2:  much descriptive info printed, and error messages printed 
 Inputs:
	the level.  
 Outputs:
	none.
 Function Returns:
        none.
*/
#ifdef __cplusplus
}
#endif
